# Copyright 2006-2009 Michel Casabianca <michel.casabianca@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

require 'bee'
require 'bee_task'
require 'bee_util'
require 'getoptlong'
require 'yaml'

# Module for Bee stuff.
module Bee

  module Console

    # Command line help.
    HELP = 'Usage: bee [options] [targets]
-V             Print version and exit.
-h             Print help about usage and exit.
-b             Print help about build and exit.
-n             Don\'t actually run any commands; just print them.
-k task        Print help about tasks in a package (writing "foo.?") or a 
               given one (writing "foo.bar") and exit.
-e egg         Print help about templates in a given package (writing 
               "foo.?") or a given one (writing "foo.bar") and exit.
-p name=value  Set a named property with a given value.
-t egg         Run a given egg to generate a template project.
-v             Enable verbose mode.
-s style       Define style for output (see documentation).
-f file        Build file to run (defaults to "build.yml").
-r             Look for build file recursively up in file system.
-l             Print bee logo on console.
targets        Targets to run (default target if omitted).'
    # Name for default build file.
    DEFAULT_BUILD_FILE = 'build.yml'
    # Exit value on error parsing command line
    EXIT_PARSING_CMDLINE = 1
    # Exit value on build error
    EXIT_BUILD_ERROR = 2
    # Exit value on unknown error
    EXIT_UNKNOWN_ERROR = 3
    # Bee options environment variable.
    BEE_OPT_ENV = 'BEEOPT'
    # Bee text logo (generated with http://www.network-science.de/ascii/)
    BEE_LOGO = <<'EOF'
         _                                                                 
        | |__   ___  ___                                                   
 _____  | '_ \ / _ \/ _ \  _____ _____ _____ _____ _____ _____ _____ _____ 
|_____| | |_) |  __/  __/ |_____|_____|_____|_____|_____|_____|_____|_____|
        |_.__/ \___|\___|  0.5.0                   http://bee.rubyforge.org

EOF

    # Parse command line and return parsed arguments.
    def self.parse_command_line
      version = false
      help = false
      help_build = false
      help_task = false
      help_template = false
      properties = {}
      task = nil
      dry_run = false
      template = nil
      verbose = false
      style = nil
      file = DEFAULT_BUILD_FILE
      recursive = false
      logo = false
      targets = []
      # read options in BEEOPT environment variable
      options = ENV[BEE_OPT_ENV]
      options.split(' ').reverse.each { |option| ARGV.unshift(option) } if
        options
      # parse command line arguments
      opts = GetoptLong.new(['--version', '-V', GetoptLong::NO_ARGUMENT],
                            ['--help', '-h', GetoptLong::NO_ARGUMENT],
                            ['--help-build', '-b', GetoptLong::NO_ARGUMENT],
                            ['--help-task','-k', GetoptLong::REQUIRED_ARGUMENT],
                            ['--help-template','-e', GetoptLong::REQUIRED_ARGUMENT],
                            ['--dry-run', '-n', GetoptLong::NO_ARGUMENT],
                            ['--property', '-p', GetoptLong::REQUIRED_ARGUMENT],
                            ['--template', '-t', GetoptLong::REQUIRED_ARGUMENT],
                            ['--verbose', '-v', GetoptLong::NO_ARGUMENT],
                            ['--style', '-s', GetoptLong::REQUIRED_ARGUMENT],
                            ['--file', '-f', GetoptLong::REQUIRED_ARGUMENT],
                            ['--recursive', '-r', GetoptLong::NO_ARGUMENT],
                            ['--logo', '-l', GetoptLong::NO_ARGUMENT])
      opts.each do |opt, arg|
        case opt
        when '--version'
          version = true
        when '--help'
          help = true
        when '--help-build'
          help_build = true
        when '--help-task'
          help_task = true
          task = arg
        when '--help-template'
          help_template = true
          template = arg
        when '--dry-run'
          dry_run = true
          verbose = true
        when '--property'
          name, value = parse_property(arg)
          properties[name] = value
        when '--template'
          template = arg
        when '--verbose'
          verbose = true
        when '--style'
          style = arg
        when '--file'
          file = arg
        when '--recursive'
          recursive = true
        when '--logo'
          logo = true
        end
      end
      targets = Array.new(ARGV)
      return [version, help, help_build, help_task, help_template, task,
              properties, dry_run, template, verbose, style, file, recursive,
              logo, targets]
    end

    # Parse a command line property.
    # - property: property definition as "name=value".
    # Return: name and value of the property.
    def self.parse_property(property)
      begin
        index = property.index('=')
        raise "No = sign" if not index
        name = property[0..index-1]
        value = YAML::load(property[index+1..-1])
        return name, value
      rescue
        raise "Error parsing property '#{property}': #{$!}"
      end
    end

    # Start build from command line.
    def self.start_command_line
      STDOUT.sync = true
      begin
        version, help, help_build, help_task, help_template, task,
          properties, dry_run, template, verbose, style, file, recursive,
          logo, targets = parse_command_line
      rescue
        puts "ERROR: parsing command line (type 'bee -h' for help)"
        exit(EXIT_PARSING_CMDLINE)
      end
      formatter = Formatter.new(style)
      begin
        if logo
          puts BEE_LOGO
        end
        if version
          copyright = Bee::Util::copyright
          puts copyright if copyright
        elsif help
          puts HELP
        elsif help_build
          build = Build.load(file, recursive, properties)
          build.evaluate_properties
          puts formatter.help_build(build)
        elsif help_task
          puts formatter.help_task(task)
        elsif help_template
          puts formatter.help_template(template)
        elsif template
          file = Bee::Util::find_template(template)
          listener = Listener.new(formatter, verbose)
          build = Build.load(file, false, properties)
          build.run(targets, listener, dry_run)
        else
          listener = Listener.new(formatter, verbose)
          build = Build.load(file, recursive, properties)
          build.run(targets, listener, dry_run)
        end
      rescue Bee::Util::BuildError => e
        puts "#{formatter.format_error('ERROR')}: #{$!}"
        puts e.backtrace.join("\n") if verbose
        exit(EXIT_BUILD_ERROR)
      rescue Exception => e
        puts "#{formatter.format_error('ERROR')}: #{$!}"
        puts e.backtrace.join("\n")
        exit(EXIT_UNKNOWN_ERROR)
      end
    end

    # Class to format build output on console.
    class Formatter

      include Bee::Util::BuildErrorMixin

      # List of colors.
      COLORS = [:black, :red, :green, :yellow, :blue, :magenta, :cyan, :white]
      # Foreground color codes.
      FOREGROUND_COLOR_CODES = {
        :black   => 30,
        :red     => 31,
        :green   => 32,
        :yellow  => 33,
        :blue    => 34,
        :magenta => 35,
        :cyan    => 36,
        :white   => 37
      }
      # Background color codes.
      BACKGROUND_COLOR_CODES = {
        :black   => 40,
        :red     => 41,
        :green   => 42,
        :yellow  => 43,
        :blue    => 44,
        :magenta => 45,
        :cyan    => 46,
        :white   => 47
      }
      # List of styles.
      STYLES = [:reset, :bright, :dim, :underscore, :blink, :reverse, :hidden]
      # Style codes.
      STYLE_CODES = {
        :reset      => 0,
        :bright     => 1,
        :dim        => 2,
        :underscore => 4,
        :blink      => 5,
        :reverse    => 7,
        :hidden     => 8
      }

      # Default style (supposed to work on any configuration).
      DEFAULT_STYLE = {
        :line_character => '-'
      }
      # Short style keys for command line
      SHORT_STYLE_KEYS = {
        'lc' => 'line_character',
        'll' => 'line_length',
        'ts' => 'target_style',
        'tf' => 'target_foreground',
        'tb' => 'target_background',
        'ks' => 'task_style',
        'kf' => 'task_foreground',
        'kb' => 'task_background',
        'ss' => 'success_style',
        'sf' => 'success_foreground',
        'sb' => 'success_background',
        'es' => 'error_style',
        'ef' => 'error_foreground',
        'eb' => 'error_background'
      }

      # Constructor.
      # - style: style as a Hash or a String.
      def initialize(style)
        # if style is a String, this is command line argument
        style = parse_style_from_command_line(style) if style.kind_of?(String)
        # if style is nil, set default style
        @style = style || DEFAULT_STYLE
        # set default values for keys if nil
        for key in DEFAULT_STYLE.keys
          @style[key] = @style[key] || DEFAULT_STYLE[key]
        end
      end

      # Format a target.
      # - target: target to format.
      def format_target(target)
        name = target.name
        return format_title(name)
      end

      # Format a task.
      # - task: task to format.
      def format_task(task)
        if task.kind_of?(String)
          source = task
        elsif task.kind_of?(Hash)
          if task.key?('rb')
            source = "rb: #{task['rb']}"
          else
            source = YAML::dump(task)
            source = source.sub(/---/, '')
          end
        end
        formatted = '- ' + source.strip.gsub(/\n/, "\n. ")
        styled = style(formatted,
                       @style[:task_style], 
                       @style[:task_foreground],
                       @style[:task_background])
        return styled
      end

      # Format a success string.
      # - string: string to format.
      def format_success(string)
        string = style(string,
                       @style[:success_style], 
                       @style[:success_foreground],
                       @style[:success_background])
        return string
      end

      # Format an error string.
      # - string: string to format.
      def format_error(string)
        string = style(string,
                       @style[:error_style], 
                       @style[:error_foreground],
                       @style[:error_background])
        return string
      end

      # Return help about build.
      # - build: running build.
      def help_build(build)
        help = ''
        # print build name and description
        if build.name
          help << "- Build: #{build.name}\n"
        end
        if build.extends
          help << "  Extends: #{build.extends}\n"
        end
        if build.description
          help << format_description('Description', build.description, 2, false)
        end
        # print build properties
        if build.context.properties.length > 0
          help << "- Properties:\n"
          for property in build.context.properties.sort
            help << "  - #{property}: " +
              "#{build.context.get_property(property, true).inspect}\n"
          end
        end
        # print build targets
        description = build.targets.description
        if description.length > 0
          help << "- Targets:\n"
          for name in description.keys.sort
            help << format_description(name, description[name], 2)
          end
        end
        # print default target
        help << "- Default: #{build.default}\n"
        return help.strip
      end

      # Return help about task(s).
      # - task: task to print help about (all tasks if nil).
      def help_task(task)
        task = '?' if task == nil or task.length == 0
        package_manager = Bee::Task::PackageManager.new(nil)
        methods = package_manager.help_task(task)
        help = ''
        for method in methods.keys.sort
          text = methods[method].strip
          help << format_title(method)
          help << "\n"
          help << text
          help << "\n"
          if text =~ /Alias for \w+/
            alias_method = text.scan(/Alias for (\w+)/).flatten[0]
            help << "\n"
            help << package_manager.help_task(alias_method)[alias_method].strip
            help << "\n"
          end
        end
        return help
      end

      # Return help about template(s).
      # - template: template to print help about (all templates if nil).
      def help_template(template)
        templates = Bee::Util::search_templates(template)
        help = ''
        for name in templates.keys
          build = YAML::load(File.read(templates[name]))
          properties = nil
          for entry in build
            properties = entry['properties'] if entry['properties']
          end
          description = 'No description found'
          if properties
            for property in properties
              if property.keys == ['description']
                description = property['description']
              end
            end
          end
          help << format_title(name)
          help << "\n"
          help << description
          help << "\n"
        end
        return help
      end

      private

      # Apply style to a string:
      # - string: the string to apply style to.
      # - style: style to apply on string.
      # - foreground: foreground color for string.
      # - background: background color for string.
      def style(string, style, foreground, background)
        # check style, foreground and background colors
        error "Unknown style '#{style}'" unless 
          STYLES.member?(style) or not style
        error "Unknown color '#{foreground}'" unless
          COLORS.member?(foreground) or not foreground
        error "Unknown color '#{background}'" unless
          COLORS.member?(background) or not background
        # if no style nor colors, return raw string
        return string if not foreground and not background and not style
        # insert style and colors in string
        colorized = "\e["
        colorized << "#{STYLE_CODES[style]};" if style
        colorized << "#{FOREGROUND_COLOR_CODES[foreground]};" if foreground
        colorized << "#{BACKGROUND_COLOR_CODES[background]};" if background
        colorized = colorized[0..-2]
        colorized << "m#{string}\e[#{STYLE_CODES[:reset]}m"
        return colorized
      end

      # Parse style from command line. If error occurs parsing style, return
      # nil (which means default style).
      # - string: style to parse.
      def parse_style_from_command_line(string)
        return if not string
        style = {}
        begin
          for pair in string.split(',')
            key, value = pair.split(':')
            key = SHORT_STYLE_KEYS[key] || key
            key = key.to_sym
            if key == :line_length
              value = value.to_i
            elsif key == :line_character
              value = ' ' if not value or value.length == 0
            else
              value = value.to_sym
              error "Unkown color or style '#{value}'" if 
                not COLORS.member?(value) and not STYLES.member?(value)
            end
            style[key] = value
          end
          return style
        rescue
          # if parsing fails, return default style (nil)
          return nil
        end
      end

      # Format a description.
      # - title: description title (project, property or target name).
      # - text: description text.
      # - indent: indentation width.
      # - bullet: tells if we must put a bullet.
      def format_description(title, text=nil, indent=0, bullet=true)
        string = ' '*indent
        string << '- ' if bullet
        string << title
        if text and !text.empty?
          string << ": "
          if text.split("\n").length > 1
            string << "\n"
            text.split("\n").each do |line|
              string << ' '*(indent+2) + line.strip + "\n"
            end
          else
            string << text.strip + "\n"
          end
        else
          string << "\n"
        end
        return string
      end

      # Format a title.
      # - title: title to format.
      def format_title(title)
        length = @style[:line_length] || Bee::Util::term_width
        right = ' ' + @style[:line_character]*2
        size = length - (title.length + 4)
        size = 2 if size <= 0
        left = @style[:line_character]*size + ' '
        line = left + title + right
        # apply style
        formatted = style(line, 
                          @style[:target_style], 
                          @style[:target_foreground],
                          @style[:target_background])
        return formatted
      end

    end

    # Listener when running in a console. Prints messages on the console using
    # a given formatter.
    class Listener

      # Formatter used by listener.
      attr_reader :formatter
      # Verbosity flag.
      attr_reader :verbose
      # Build start time.
      attr_reader :start_time
      # Build end time.
      attr_reader :end_time
      # Build duration.
      attr_reader :duration
      # Build success.
      attr_reader :success
      # Last target met.
      attr_reader :last_target
      # Last task met.
      attr_reader :last_task

      # Constructor.
      # - formatter: the formatter to use to output on console.
      # - verbose: tells if we run in verbose mode.
      def initialize(formatter, verbose)
        @formatter = formatter
        @verbose = verbose
      end

      # Called when build is started.
      # - build: the build object.
      def build_started(build, dry_run)
        @start_time = Time.now
        @end_time = nil
        @duration = nil
        @success = nil
        @last_target = nil
        @last_task = nil
        build_type = dry_run ? "dry run of" : "build"
        puts "Starting #{build_type} '#{build.file}'..." if @verbose
      end

      # Called when a target is met.
      # - target: the target object.
      def target(target)
        @last_target = target
        @last_task = nil
        puts @formatter.format_target(target)
      end

      # Called when a task is met.
      # - task: task source (shell, Ruby or task).
      def task(task)
        @last_task = task
        puts @formatter.format_task(task) if @verbose
      end

      # Called when build is finished.
      # - build: the build object.
      def build_finished(build, dry_run)
        duration unless dry_run
        @success = true
        puts @formatter.format_success('OK')
      end

      # Called when an error was raised.
      # - exception: raised exception.
      def error(exception)
        duration
        @success = false
        message = ''
        message << "In target '#{@last_target.name}'" if @last_target
        message << ", in task:\n#{@formatter.format_task(@last_task)}\n" if
          @last_task
        message << ': ' if @last_target and not @last_task
        message << exception.to_s
        raise Bee::Util::BuildError.new(message)
      end

      # Print text on the console.
      # - text: text to print.
      def print(text)
        Kernel.print(text)
      end
      
      # Puts text on the console.
      # - text: text to puts.
      def puts(text)
        Kernel.puts(text)
      end

      # Get input string.
      # Return string entered by the user.
      def gets
        return STDIN.gets
      end

      private

      # Compute build duration and print it if verbose.
      def duration
        @end_time = Time.now
        @duration = @end_time - @start_time
        puts "Built in #{@duration} s" if @verbose        
      end
      
    end

  end

end
